import logging
import sys
import time

from influxdb_client import InfluxDBClient

logging.basicConfig(
    stream=sys.stdout,
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


class Loader:
    """
    Base class for loading IB stats into InfluxDB
    """

    def __init__(self, config):
        # Configure logging
        self.log = logging.getLogger("ib2influx")

        # Collection interval in seconds
        self.interval = config["interval"]

        self.influx_config = config["influxdb"]

        self.log.info(
            f"Writing to {self.influx_config['server']} at {self.interval}s intervals"
        )

        # Set up InfluxDB client
        self.influx_client = InfluxDBClient(
            url=self.influx_config["server"],
            token=self.influx_config["token"],
            org=self.influx_config["org"],
        )

        self.run()

    def run(self):
        """
        Collect statistics and write to InfluxDB at regular time intervals
        """

        while True:
            time_start = self.timestamp()

            self.influx_write(self.data_dict())

            time_finish = self.timestamp()

            dt = time_finish - time_start
            sleep_time = max(0, self.interval - dt)
            time.sleep(sleep_time)

    def influx_write(self, data):
        """
        Write data dictionary to InfluxDB
        """
        write_api = self.influx_client.write_api()

        # Dictionary-style write for each item in dict
        for host in data:
            entry = [
                {
                    "measurement": "infiniband",
                    "tags": {"host": host},
                    "fields": data[host],
                    "time": self.time,
                }
            ]

            write_api.write(
                self.influx_config["bucket"],
                self.influx_config["org"],
                entry,
                write_precision="s",
            )

        write_api.close()

    def set_time(self):
        """
        Sets self.time to the current time
        Used for getting the timestamp of the collected data
        """
        self.time = self.timestamp()

    @staticmethod
    def timestamp():
        """
        Return current timestamp
        """
        return int(time.time())



import os
import subprocess
from glob import glob


def cmd(command):
    """
    Run a command, checking that the return code == 0
    """
    p = subprocess.run(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
    )
    return p.stdout.decode()


def find_most_recent_file(path, suffix):
    """
    Find the most recent file in path with suffix
    """
    files = glob(os.path.join(path, "*"))
    last = None
    last_time = 0
    for f in files:
        if f.split(".")[-1] == suffix:
            m = os.path.getmtime(f)
            if m > last_time:
                last = f
                last_time = m
    return last, last_time



from .loader import Loader
from .utils import cmd, find_most_recent_file


class LoaderMLX(Loader):
    """
    Loader for Mellanox hardware
    """

    def ib_port_map(self):
        """
        Generate a host to switch port mapping
        """

        # Location of IB files
        ib_path = "/root/ib"

        # Get the most recent ibnetdiscover file
        ibnetdiscover_file, _ = find_most_recent_file(ib_path, "ibnetdiscover")

        with open(ibnetdiscover_file, "r") as f:
            lines = f.readlines()

        # Dictionary of mapping
        self.port_map = {}
        self.switch_lids = []
        self.switch_port_count = []

        for line in lines:
            items = line.split()

            if len(items) == 0:
                continue

            # If line describes a switch
            if items[0] == "Switch":
                # Exclude spine switches
                if "spine" in items[4]:
                    continue

                # Get number of ports on the switch (minus one for the base port)
                self.switch_port_count += [int(items[1]) - 1]

                # Get the switch LID
                for i, txt in enumerate(items):
                    if txt == "lid":
                        switch_lid = items[i + 1]
                        self.switch_lids += [switch_lid]
                        break

            # If the first item matches a number inside square brackets
            elif items[0][0] == "[" and items[0][-1] == "]":
                # Get the port number
                port = items[0][1:-1]
                # Get the host name
                host = items[3]

                # If not a spine switch
                if "spine" in host:
                    continue

                # If not a port on a spine switch
                if "Quantum-2" in host:
                    continue

                self.port_map[(switch_lid, port)] = host.lstrip('"')

    def query_switches(self):
        """
        Query a switch for performance data
        """
        self.set_time()

        data = {}
        host = None

        # Loop over switches
        for lid, n in zip(self.switch_lids, self.switch_port_count):
            output = cmd(f"perfquery -x -l {lid} 1-{n}")

            for line in output.split("\n"):
                if "PortSelect" in line:
                    # Get the number at the end of line separated by "."
                    port = line.split(".")[-1]

                    # Look up host in dict, default to None
                    host = self.port_map.get((lid, port), None)
                    # Host variable is now set for reading subsequent values
                    if host is not None:
                        data[host] = {}

                # Read values only if this is an actual host
                if host is not None:
                    # Note: Xmit = in and Rcv = out because this is on the switch side
                    # Counters are in units of 32 bits, so multiply by 4 to get bytes

                    if "PortXmitData" in line:
                        data[host]["bytes_in"] = int(line.split(".")[-1]) * 4

                    elif "PortRcvData" in line:
                        data[host]["bytes_out"] = int(line.split(".")[-1]) * 4

                    elif "PortXmitPkts" in line:
                        data[host]["pkts_in"] = int(line.split(".")[-1])

                    elif "PortRcvPkts" in line:
                        data[host]["pkts_out"] = int(line.split(".")[-1])

        return data

    def data_dict(self):
        self.ib_port_map()
        return self.query_switches()



import shutil
import sys
from pathlib import Path

import yaml

from . import mlx, opa


def run(argv=sys.argv):
    if len(argv) < 2:
        print("Usage: ib2influx config.yml")
    else:
        config_file = Path(argv[1])
        if config_file.is_file():
            with open(config_file, "r") as f:
                config = yaml.safe_load(f)
        else:
            raise FileNotFoundError

        # Check if we are using OPA by checking if the opaextractperf executable is in the path
        path = shutil.which("opaextractperf")
        if path is not None:
            print("OPA found")
            opa.LoaderOPA(config)

        # Check if we are using Mellanox by checking if the ibnetdiscover executable is in the path
        path = shutil.which("ibnetdiscover")
        if path is not None:
            print("Mellanox found")
            mlx.LoaderMLX(config)
